from abc import ABCMeta, abstractmethod
from typing import Optional, Text

import attr
from six import add_metaclass

from applitools.common.geometry import RectangleSize
from applitools.common.selenium.misc import BrowserType
from applitools.common.utils.json_utils import JsonInclude

from .config import DeviceName, IosDeviceName, IosVersion, ScreenOrientation


@add_metaclass(ABCMeta)
class IRenderBrowserInfo(object):
    @property
    @abstractmethod
    def platform(self):
        # type: () -> Text
        pass

    @property
    @abstractmethod
    def browser(self):
        # type: () -> Text
        pass

    @property
    @abstractmethod
    def width(self):
        # type: () -> int
        pass

    @property
    @abstractmethod
    def height(self):
        # type: () -> int
        pass

    @property
    def viewport_size(self):
        # type: () -> RectangleSize
        return RectangleSize(self.width, self.height)

    @property
    def baseline_env_name(self):
        # type: () -> Optional[Text]
        return self._baseline_env_name


@attr.s
class EmulationBaseInfo(object):
    device_name = attr.ib()  # type: DeviceName
    screen_orientation = attr.ib()  # type: ScreenOrientation


@attr.s(hash=True, init=False)
class ChromeEmulationInfo(EmulationBaseInfo, IRenderBrowserInfo):
    device_name = attr.ib(
        type=DeviceName, metadata={JsonInclude.THIS: True}
    )  # type: DeviceName
    screen_orientation = attr.ib(
        type=ScreenOrientation, metadata={JsonInclude.NON_NONE: True}
    )  # type: ScreenOrientation

    def __init__(
        self,
        device_name,  # type: DeviceName
        screen_orientation=ScreenOrientation.PORTRAIT,  # type: ScreenOrientation
        baseline_env_name=None,  # type: Optional[Text]
    ):
        # type: (...) -> None
        self.device_name = DeviceName(device_name)
        self.screen_orientation = ScreenOrientation(screen_orientation)
        self._baseline_env_name = baseline_env_name

    @property
    def width(self):
        # type: () -> int
        return 0

    @property
    def height(self):
        # type: () -> int
        return 0

    @property
    def browser(self):
        return BrowserType.CHROME.value

    @property
    def platform(self):
        # type: () -> Text
        return "linux"


@attr.s(hash=True, init=False)
class IosDeviceInfo(IRenderBrowserInfo):
    device_name = attr.ib(
        type=IosDeviceName, metadata={JsonInclude.NAME: "name"}
    )  # type: DeviceName
    screen_orientation = attr.ib(
        type=ScreenOrientation, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[ScreenOrientation]
    ios_version = attr.ib(
        type=IosVersion, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[IosVersion]

    def __init__(
        self,
        device_name,  # type: IosDeviceName
        screen_orientation=ScreenOrientation.PORTRAIT,  # type: ScreenOrientation
        ios_version=None,  # type: Optional[IosVersion]
        baseline_env_name=None,  # type: Optional[Text]
    ):
        self.device_name = IosDeviceName(device_name)
        self.screen_orientation = ScreenOrientation(screen_orientation)
        self.ios_version = IosVersion(ios_version) if ios_version else None
        self._baseline_env_name = baseline_env_name

    @property
    def width(self):
        # type: () -> int
        return 0

    @property
    def height(self):
        # type: () -> int
        return 0

    @property
    def platform(self):
        # type: () -> Text
        return "ios"

    @property
    def browser(self):
        # type: () -> Text
        return BrowserType.SAFARI.value


@attr.s(hash=True)
class RenderBrowserInfo(IRenderBrowserInfo):
    viewport_size = attr.ib(
        default=None, converter=attr.converters.optional(RectangleSize.from_)
    )  # type: Optional[RectangleSize]  # type: ignore
    browser_type = attr.ib(
        default=BrowserType.CHROME,
        converter=attr.converters.optional(BrowserType),
        type=BrowserType,
    )  # type: BrowserType
    baseline_env_name = attr.ib(default=None)  # type: Optional[Text]
    emulation_info = attr.ib(
        default=None, repr=False
    )  # type: Optional[EmulationBaseInfo]

    @property
    def width(self):
        # type: () -> int
        if self.viewport_size:
            return self.viewport_size["width"]
        return 0

    @property
    def height(self):
        # type: () -> int
        if self.viewport_size:
            return self.viewport_size["height"]
        return 0

    @property
    def platform(self):
        # type: () -> Text
        if self.browser_type in [
            BrowserType.IE_10,
            BrowserType.IE_11,
            BrowserType.EDGE_LEGACY,
        ]:
            return "windows"
        elif self.browser_type in [
            BrowserType.SAFARI,
            BrowserType.SAFARI_ONE_VERSION_BACK,
            BrowserType.SAFARI_TWO_VERSIONS_BACK,
            BrowserType.SAFARI_EARLY_ACCESS,
        ]:
            return "mac os x"
        return "linux"

    @property
    def browser(self):
        # type: () -> Text
        return self.browser_type.value if self.browser_type else None


@attr.s(hash=True, init=False)
class DesktopBrowserInfo(IRenderBrowserInfo):
    _width = attr.ib()  # type: int
    _height = attr.ib()  # type: int
    browser_type = attr.ib(type=BrowserType)  # type: BrowserType

    def __init__(
        self, width, height, browser_type=BrowserType.CHROME, baseline_env_name=None
    ):
        # type: (int, int, BrowserType, Optional[Text])->None
        self._width = width
        self._height = height
        self._baseline_env_name = baseline_env_name
        self.browser_type = BrowserType(browser_type)

    @property
    def width(self):
        # type: () -> int
        return self._width

    @property
    def height(self):
        # type: () -> int
        return self._height

    @property
    def browser(self):
        # type: () -> Text
        return self.browser_type.value

    @property
    def platform(self):
        # type: () -> Text
        if self.browser_type in [
            BrowserType.IE_10,
            BrowserType.IE_11,
            BrowserType.EDGE_LEGACY,
        ]:
            return "windows"
        elif self.browser_type in [
            BrowserType.SAFARI,
            BrowserType.SAFARI_ONE_VERSION_BACK,
            BrowserType.SAFARI_TWO_VERSIONS_BACK,
            BrowserType.SAFARI_EARLY_ACCESS,
        ]:
            return "mac os x"
        return "linux"



from __future__ import absolute_import

from enum import Enum
from os import getcwd
from threading import Lock
from typing import Any, List, Optional, Text

from selenium.common.exceptions import StaleElementReferenceException

from ..common.errors import USDKFailure
from .connection import USDKConnection

# A backward-compatible alias, Exception was named Failure in original 5.0.0 release
Failure = USDKFailure


class ManagerType(Enum):
    VG = "vg"
    CLASSIC = "classic"


class CommandExecutor(object):
    @classmethod
    def create(cls, name, version):
        # type: (Text, Text) -> CommandExecutor
        commands = cls(USDKConnection.create())
        commands.make_sdk(name, version, getcwd())
        return commands

    @classmethod
    def get_instance(cls, name, version):
        # type: (Text, Text) -> CommandExecutor
        with _instances_lock:
            key = (name, version)
            if key in _instances:
                return _instances[key]
            else:
                return _instances.setdefault(key, cls.create(name, version))

    def __init__(self, connection):
        # type: (USDKConnection) -> None
        self._connection = connection

    def make_sdk(self, name, version, cwd):
        # type: (Text, Text, Text) -> None
        self._connection.notification(
            "Core.makeSDK",
            {"name": name, "version": version, "cwd": cwd, "protocol": "webdriver"},
        )

    def core_make_manager(self, manager_type, concurrency=None, is_legacy=None):
        # type: (ManagerType, Optional[int], Optional[bool]) -> dict
        payload = {"type": manager_type.value}
        if concurrency is not None:
            payload["concurrency"] = concurrency
        if is_legacy is not None:
            payload["legacy"] = is_legacy
        return self._checked_command("Core.makeManager", payload)

    def core_get_viewport_size(self, driver):
        # type: (dict) -> dict
        return self._checked_command("Core.getViewportSize", {"driver": driver})

    def core_set_viewport_size(self, driver, size):
        # type: (dict, dict) -> None
        self._checked_command("Core.setViewportSize", {"driver": driver, "size": size})

    def core_close_batches(self, close_batches_settings):
        # type: (dict) -> None
        self._checked_command("Core.closeBatches", {"settings": close_batches_settings})

    def core_delete_test(self, close_test_settings):
        # type: (dict) -> None
        self._checked_command("Core.deleteTest", {"settings": close_test_settings})

    def manager_open_eyes(self, manager, driver, config=None):
        # type: (dict, dict, Optional[dict]) -> dict
        payload = {"manager": manager, "driver": driver}
        if config is not None:
            payload["config"] = config
        return self._checked_command("EyesManager.openEyes", payload)

    def manager_close_all_eyes(self, manager, timeout):
        # type: (dict, float) -> List[dict]
        return self._checked_command(
            "EyesManager.closeAllEyes", {"manager": manager}, wait_timeout=timeout
        )

    def eyes_check(self, eyes, settings=None, config=None):
        # type: (dict, Optional[dict], Optional[dict]) -> dict
        payload = {"eyes": eyes}
        if settings is not None:
            payload["settings"] = settings
        if config is not None:
            payload["config"] = config
        return self._checked_command("Eyes.check", payload)

    def eyes_locate(self, eyes, settings, config=None):
        # type: (dict, dict, Optional[dict]) -> dict
        payload = {"eyes": eyes, "settings": settings}
        if config:
            payload["config"] = config
        return self._checked_command("Eyes.locate", payload)

    def eyes_extract_text(self, eyes, regions, config=None):
        # type: (dict, dict, Optional[dict]) -> List[Text]
        payload = {"eyes": eyes, "regions": regions}
        if config:
            payload["config"] = config
        return self._checked_command("Eyes.extractText", payload)

    def eyes_extract_text_regions(self, eyes, settings, config=None):
        # type: (dict, dict, Optional[dict]) -> dict
        payload = {"eyes": eyes, "settings": settings}
        if config:
            payload["config"] = config
        return self._checked_command("Eyes.extractTextRegions", payload)

    def eyes_close_eyes(self, eyes, wait_result):
        # type: (dict, bool) -> List[dict]
        return self._checked_command("Eyes.close", {"eyes": eyes}, wait_result)

    def eyes_abort_eyes(self, eyes, wait_result):
        # type: (dict, bool) -> List[dict]
        return self._checked_command("Eyes.abort", {"eyes": eyes}, wait_result)

    def server_get_info(self):
        # type: () -> dict
        return self._checked_command("Server.getInfo", {})

    def _checked_command(self, name, payload, wait_result=True, wait_timeout=9 * 60):
        # type: (Text, dict, bool, float) -> Optional[Any]
        response = self._connection.command(name, payload, wait_result, wait_timeout)
        if wait_result:
            response_payload = response["payload"]
            _check_error(response_payload)
            return response_payload.get("result")
        else:
            return None


def _check_error(payload):
    # type: (dict) -> None
    error = payload.get("error")
    if error:
        if error["message"].startswith("stale element reference"):
            raise StaleElementReferenceException(error["message"])
        else:
            raise USDKFailure(error["message"], error["stack"])


_instances = {}
_instances_lock = Lock()



from json import dumps
from typing import TYPE_CHECKING, Iterable, Text, Tuple, Union

import attr
import cattr
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from six import text_type

from applitools.common import (
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
)
from applitools.common import ChromeEmulationInfo as ApiChromeEmulationInfo
from applitools.common import DesktopBrowserInfo, DeviceName, FloatingBounds
from applitools.common import IosDeviceInfo
from applitools.common import IosDeviceInfo as ApiIosDeviceInfo
from applitools.common import IosDeviceName, IosVersion, MatchLevel, MatchResult
from applitools.common import Region as APIRegion
from applitools.common import (
    RenderBrowserInfo,
    ScreenOrientation,
    SessionType,
    StitchMode,
    TestResults,
    VisualGridOption,
)
from applitools.common.utils.json_utils import attr_from_json, underscore_to_camelcase

from ..common.geometry import Rectangle
from ..core import (
    FloatingRegionByRectangle,
    GetRegion,
    RegionByRectangle,
    TextRegionSettings,
    VisualLocatorSettings,
)
from ..core.fluent import AccessibilityRegionByRectangle
from ..core.locators import VisualLocatorSettingsValues
from .fluent import (
    FloatingRegionByElement,
    FloatingRegionBySelector,
    RegionByElement,
    RegionBySelector,
)
from .fluent.region import AccessibilityRegionByElement, AccessibilityRegionBySelector
from .fluent.target_path import FrameLocator, Locator, RegionLocator, ShadowDomLocator

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Union

    from selenium.webdriver.remote.webdriver import WebDriver, WebElement

    from applitools.common import BatchInfo, ImageMatchSettings, ProxySettings
    from applitools.common.selenium import BrowserType, Configuration
    from applitools.core.batch_close import _EnabledBatchClose

    from ..common.utils.custom_types import ViewPort
    from . import OCRRegion
    from .fluent import FrameLocator, SeleniumCheckSettings
    from .fluent.selenium_check_settings import SeleniumCheckSettingsValues


@attr.s
class CustomProperty(object):
    name = attr.ib()  # type: Text
    value = attr.ib()  # type: Text

    @classmethod
    def convert(cls, properties):
        # type: (List[Dict[Text, Text]]) -> List[CustomProperty]
        if properties:
            return [cls(p["name"], p["value"]) for p in properties]
        else:
            return []


@attr.s
class Batch(object):
    id = attr.ib(default=None)  # type: Optional[Text]
    name = attr.ib(default=None)  # type: Optional[Text]
    sequence_name = attr.ib(default=None)  # type: Optional[Text]
    started_at = attr.ib(default=None)  # type: Optional[Text]
    notify_on_completion = attr.ib(default=None)  # type: Optional[bool]
    properties = attr.ib(default=None)  # type: Optional[List[CustomProperty]]

    @classmethod
    def convert(cls, batch_info):
        # type: (Optional[BatchInfo]) -> Optional[Batch]
        if batch_info:
            if batch_info.started_at:
                started_at = batch_info.started_at.strftime("%Y-%m-%dT%H:%M:%SZ")
            else:
                started_at = None
            return cls(
                batch_info.id,
                batch_info.name,
                batch_info.sequence_name,
                started_at,
                batch_info.notify_on_completion,
                # TODO: Verify
                CustomProperty.convert(batch_info.properties),
            )
        else:
            return None


@attr.s
class Location(object):
    x = attr.ib()  # type: float
    y = attr.ib()  # type: float


@attr.s
class Size(object):
    width = attr.ib()  # type: float
    height = attr.ib()  # type: float

    @classmethod
    def convert_viewport(cls, viewport):
        # type: (Optional[ViewPort]) -> Optional[Size]
        if viewport:
            return cls(viewport["width"], viewport["height"])
        else:
            return None


@attr.s
class Region(Location, Size):
    @classmethod
    def convert(cls, rect):
        # type: (Union[Region, Rectangle]) -> Region
        return cls(rect.left, rect.top, rect.width, rect.height)


@attr.s
class ImageCropRegion(Region):
    pass


@attr.s
class ImageCropRect(object):
    top = attr.ib()  # type: float
    right = attr.ib()  # type: float
    bottom = attr.ib()  # type: float
    left = attr.ib()  # type: float


@attr.s
class AccessibilityRegion(object):
    region = attr.ib()  # type: Region
    type = attr.ib(default=None)  # type: Optional[AccessibilityRegionType]

    @classmethod
    def convert(cls, region, type_):
        # type: (Region, Optional[AccessibilityRegionType]) -> AccessibilityRegion
        return cls(region, type_)


@attr.s
class FloatingRegion(object):
    region = attr.ib()  # type: Region
    max_up_offset = attr.ib()  # type: Optional[int]
    max_down_offset = attr.ib()  # type: Optional[int]
    max_left_offset = attr.ib()  # type: Optional[int]
    max_right_offset = attr.ib()  # type: Optional[int]

    @classmethod
    def convert(cls, region, bounds):
        # type: (Region, FloatingBounds) -> FloatingRegion
        return cls(
            region,
            bounds.max_up_offset,
            bounds.max_down_offset,
            bounds.max_left_offset,
            bounds.max_right_offset,
        )


@attr.s
class TransformedElement(object):
    element_id = attr.ib(type=text_type)

    @classmethod
    def convert(cls, web_element):
        # type: (Optional[WebElement]) -> Optional[TransformedElement]
        if web_element:
            return TransformedElement(web_element._id)  # noqa
        else:
            return None


@attr.s
class TransformedSelector(object):
    selector = attr.ib()  # type: Union[Text, TransformedSelector]
    type = attr.ib()  # type: Optional[Text]
    shadow = attr.ib()  # type: Union[Text, TransformedSelector]
    frame = attr.ib()  # type: Union[Text, TransformedSelector]

    @classmethod
    def from_by(cls, is_selenium, selector, by=None, shadow=None, frame=None):
        if is_selenium:
            if by == By.ID:
                by = By.CSS_SELECTOR
                selector = '[id="{}"]'.format(selector)
            elif by == By.TAG_NAME:
                by = By.CSS_SELECTOR
            elif by == By.CLASS_NAME:
                by = By.CSS_SELECTOR
                selector = "." + selector
            elif by == By.NAME:
                by = By.CSS_SELECTOR
                selector = '[name="{}"]'.format(selector)
        return cls(selector, by, shadow, frame)

    @classmethod
    def convert(cls, is_selenium, locator):
        # type: (bool, Locator) -> TransformedSelector
        if locator is None:
            return None
        stack = []
        while locator:
            stack.append(locator)
            locator = locator.parent
        return cls._convert_list(is_selenium, stack)

    @classmethod
    def _convert_list(cls, is_selenium, stack):
        if stack:
            head = stack.pop()
            if isinstance(head, RegionLocator):
                return cls.from_by(is_selenium, head.selector, head.by)
            elif isinstance(head, ShadowDomLocator):
                return cls.from_by(
                    is_selenium,
                    head.selector,
                    head.by,
                    cls._convert_list(is_selenium, stack),
                )
            elif isinstance(head, FrameLocator):
                return cls.from_by(
                    is_selenium,
                    head.selector,
                    head.by,
                    frame=cls._convert_list(is_selenium, stack),
                )

            else:
                raise TypeError("Unexpected Locator type", type(head))
        else:
            return None


@attr.s
class MatchSettingsExact(object):
    min_diff_intensity = attr.ib()  # type: int
    min_diff_width = attr.ib()  # type: int
    min_diff_height = attr.ib()  # type: int
    match_threshold = attr.ib()  # type: int


@attr.s
class AccessibilitySettings(object):
    level = attr.ib(
        default=None, type=AccessibilityLevel
    )  # type: Optional[AccessibilityLevel]
    guidelines_version = attr.ib(
        default=None, type=AccessibilityGuidelinesVersion
    )  # type: Optional[AccessibilityGuidelinesVersion]


@attr.s
class DesktopBrowserRenderer(object):
    width = attr.ib()  # type: float
    height = attr.ib()  # type: float
    name = attr.ib(default=None)  # type: Optional[BrowserType]


@attr.s
class ChromeEmulationInfo(object):
    device_name = attr.ib()  # type: DeviceName
    screen_orientation = attr.ib(default=None)  # type: Optional[ScreenOrientation]


@attr.s
class ChromeEmulationDeviceRenderer(object):
    chrome_emulation_info = attr.ib()  # type: ChromeEmulationInfo


@attr.s
class IosDeviceInfo(object):
    device_name = attr.ib()  # type: IosDeviceName
    ios_version = attr.ib(default=None)  # type: Optional[IosVersion]
    screen_orientation = attr.ib(default=None)  # type: Optional[ScreenOrientation]


@attr.s
class IOSDeviceRenderer(object):
    ios_device_info = attr.ib()  # type: IosDeviceInfo


@attr.s
class FileLogHandler(object):
    type = attr.ib(default="file")
    filename = attr.ib(default=None)  # type: Optional[Text]
    append = attr.ib(default=None)  # type: Optional[bool]


@attr.s
class ConsoleLogHandler(object):
    type = attr.ib(default="console")


@attr.s
class DebugScreenshotHandler(object):
    save = attr.ib()  # type: bool
    path = attr.ib(default=None)  # type: Optional[Text]
    prefix = attr.ib(default=None)  # type: Optional[Text]


LogHandler = Union[FileLogHandler, ConsoleLogHandler]
ElementReference = Union[TransformedElement, TransformedSelector]
RegionReference = Union[ElementReference, Region]
FrameReference = Union[ElementReference, int, Text]
BrowserInfo = Union[
    DesktopBrowserRenderer, ChromeEmulationDeviceRenderer, IOSDeviceRenderer
]


def record_convert(records):
    # type: (Optional[Iterable[VisualGridOption]]) -> Optional[Dict[Text, Any]]
    if records:
        return {r.key: r.value for r in records}
    else:
        return None


def element_reference_convert(is_selenium, selector=None, element=None):
    # type: (bool, Optional[Locator], Optional[WebElement]) -> ElementReference
    if selector is not None:
        return TransformedSelector.convert(is_selenium, selector)
    elif element is not None:
        return TransformedElement.convert(element)
    else:
        return None


def frame_reference_convert(
    is_selenium, selector=None, element=None, number=None, name=None
):
    # type: (bool, Optional[List[Text, Text]], WebElement, int, Text) -> FrameReference
    if name is not None:
        return name
    elif number is not None:
        return number
    else:
        return element_reference_convert(is_selenium, selector, element)


def browsers_info_convert(browsers_info):
    # type: (List[RenderBrowserInfo]) -> List[BrowserInfo]
    result = []
    for bi in browsers_info:
        if isinstance(bi, DesktopBrowserInfo):
            result.append(DesktopBrowserRenderer(bi.width, bi.height, bi.browser))
        elif isinstance(bi, ApiChromeEmulationInfo):
            result.append(
                ChromeEmulationDeviceRenderer(
                    ChromeEmulationInfo(bi.device_name, bi.screen_orientation)
                )
            )
        elif isinstance(bi, ApiIosDeviceInfo):
            result.append(
                IOSDeviceRenderer(
                    IosDeviceInfo(bi.device_name, bi.ios_version, bi.screen_orientation)
                )
            )
        else:
            raise RuntimeError("Unexpected BrowserInfo type", type(bi))
    return result


def target_reference_convert(is_selenium, values):
    # type: (bool, SeleniumCheckSettingsValues) -> Optional[RegionReference]
    if values.target_selector:
        return TransformedSelector.convert(is_selenium, values.target_selector)
    elif values.target_element:
        return TransformedElement.convert(values.target_element)
    elif values.target_region:
        return Region.convert(values.target_region)
    else:
        return None


def ocr_target_convert(is_selenium, target):
    # type:(bool, Union[Locator, WebElement, Region]) -> RegionReference
    if isinstance(target, Locator):
        return TransformedSelector.convert(is_selenium, target)
    elif isinstance(target, WebElement):
        return TransformedElement.convert(target)
    else:
        return Region.convert(target)


def region_references_convert(is_selenium, regions):
    # type: (bool, List[GetRegion]) -> List[RegionReference]
    results = []
    for r in regions:
        if isinstance(r, RegionByElement):
            results.append(TransformedElement.convert(r._element))  # noqa
        elif isinstance(r, RegionBySelector):
            results.append(
                TransformedSelector.convert(is_selenium, r._target_path)  # noqa
            )
        elif isinstance(r, RegionByRectangle):
            results.append(Region.convert(r._region))  # noqa
        else:
            raise RuntimeError("Unexpected region type", type(r))
    return results


def floating_region_references_convert(is_selenium, regions):
    # type: (bool, List[GetRegion]) -> List[FloatingRegion]
    results = []
    for r in regions:
        if isinstance(r, FloatingRegionByElement):
            region = TransformedElement.convert(r._element)  # noqa
            bounds = r._bounds  # noqa
        if isinstance(r, FloatingRegionBySelector):
            region = TransformedSelector.convert(is_selenium, r._target_path)  # noqa
            bounds = r._bounds  # noqa
        elif isinstance(r, FloatingRegionByRectangle):
            region, bounds = Region.convert(r._rect), r._bounds  # noqa
        else:
            raise RuntimeError("Unexpected region type", type(r))
        results.append(FloatingRegion.convert(region, bounds))
    return results


def accessibility_region_references_convert(is_selenium, regions):
    # type: (bool, List[GetRegion]) -> List[AccessibilityRegion]
    results = []
    for r in regions:
        if isinstance(r, AccessibilityRegionByElement):
            region = TransformedElement.convert(r._element)  # noqa
            type_ = r._type  # noqa
        if isinstance(r, AccessibilityRegionBySelector):
            region = TransformedSelector.convert(is_selenium, r._target_path)  # noqa
            type_ = r._type  # noqa
        elif isinstance(r, AccessibilityRegionByRectangle):
            region, type_ = Region.convert(r._rect), r._type  # noqa
        else:
            raise RuntimeError("Unexpected region type", type(r))
        results.append(AccessibilityRegion.convert(region, type_))
    return results


@attr.s
class MatchSettings(object):
    exact = attr.ib(default=None)  # type: Optional[MatchSettingsExact]
    match_level = attr.ib(default=None)  # type: Optional[MatchLevel]
    send_dom = attr.ib(default=None)  # type: Optional[bool]
    use_dom = attr.ib(default=None)  # type: Optional[bool]
    enable_patterns = attr.ib(default=None)  # type: Optional[bool]
    ignore_caret = attr.ib(default=None)  # type: Optional[bool]
    ignore_displacements = attr.ib(default=None)  # type: Optional[bool]
    accessibility_settings = attr.ib(
        default=None
    )  # type: Optional[AccessibilitySettings]
    ignore_regions = attr.ib(default=None)  # type: Optional[List[RegionReference]]
    layout_regions = attr.ib(default=None)  # type: Optional[List[RegionReference]]
    strict_regions = attr.ib(default=None)  # type: Optional[List[RegionReference]]
    content_regions = attr.ib(default=None)  # type: Optional[List[RegionReference]]
    floating_regions = attr.ib(
        default=None
    )  # type: Optional[List[Union[Region, FloatingRegion]]]
    accessibility_regions = attr.ib(
        default=None
    )  # type: Optional[List[Union[Region, AccessibilityRegion]]]

    @classmethod
    def convert(cls, is_selenium, image_match_settings):
        # type: (bool, Optional[ImageMatchSettings]) -> Optional[MatchSettings]
        if image_match_settings:
            return cls(
                exact=image_match_settings.exact,
                match_level=image_match_settings.match_level,
                send_dom=None,  # TODO: verify
                use_dom=image_match_settings.use_dom,
                enable_patterns=image_match_settings.enable_patterns,
                ignore_caret=image_match_settings.ignore_caret,
                ignore_displacements=image_match_settings.ignore_displacements,
                accessibility_settings=image_match_settings.accessibility_settings,
                ignore_regions=image_match_settings.ignore_regions,  # todo: verify
                layout_regions=image_match_settings.layout_regions,
                strict_regions=image_match_settings.strict_regions,
                content_regions=image_match_settings.content_regions,
                floating_regions=floating_region_references_convert(
                    is_selenium, image_match_settings.floating_match_settings
                ),
                accessibility_regions=accessibility_region_references_convert(
                    is_selenium, image_match_settings.accessibility
                ),
            )
        else:
            return None


@attr.s
class Proxy(object):
    url = attr.ib()  # type: Text
    username = attr.ib(default=None)  # type: Optional[Text]
    password = attr.ib(default=None)  # type: Optional[Text]
    is_http_only = attr.ib(default=None)  # type: Optional[bool]

    @classmethod
    def convert(cls, proxy):
        # type: (Optional[ProxySettings]) -> Proxy
        if proxy:
            return cls(proxy.url, proxy.username, proxy.password)
        else:
            return None


@attr.s
class RemoteEvents(object):
    server_url = attr.ib()  # type: Text
    access_key = attr.ib(default=None)  # type: Optional[Text]
    timeout = attr.ib(default=None)  # type: Optional[int]


@attr.s
class EyesBaseConfig(object):
    logs = attr.ib(default=None)  # type: Optional[LogHandler]
    debug_screenshots = attr.ib(default=None)  # type: Optional[DebugScreenshotHandler]
    agent_id = attr.ib(default=None)  # type: Optional[Text]
    api_key = attr.ib(default=None)  # type: Optional[Text]
    server_url = attr.ib(default=None)  # type: Optional[Text]
    proxy = attr.ib(default=None)  # type: Optional[Proxy]
    is_disabled = attr.ib(default=None)  # type: Optional[bool]
    connection_timeout = attr.ib(default=None)  # type: Optional[int]
    remove_session = attr.ib(default=None)  # type: Optional[bool]
    remote_events = attr.ib(default=None)  # type: Optional[RemoteEvents]


@attr.s
class EyesOpenConfig(object):
    app_name = attr.ib(default=None)  # type: Optional[Text]
    test_name = attr.ib(default=None)  # type: Optional[Text]
    display_name = attr.ib(default=None)  # type: Optional[Text]
    viewport_size = attr.ib(default=None)  # type: Optional[Size]
    session_type = attr.ib(default=None)  # type: Optional[SessionType]
    properties = attr.ib(default=None)  # type: Optional[List[CustomProperty]]
    batch = attr.ib(default=None)  # type: Optional[Batch]
    default_match_settings = attr.ib(default=None)  # type: Optional[MatchSettings]
    host_app = attr.ib(default=None)  # type: Optional[Text]
    host_o_s = attr.ib(default=None)  # type: Optional[Text]
    host_app_info = attr.ib(default=None)  # type: Optional[Text]
    host_o_s_info = attr.ib(default=None)  # type: Optional[Text]
    device_info = attr.ib(default=None)  # type: Optional[Text]
    baseline_env_name = attr.ib(default=None)  # type: Optional[Text]
    environment_name = attr.ib(default=None)  # type: Optional[Text]
    branch_name = attr.ib(default=None)  # type: Optional[Text]
    parent_branch_name = attr.ib(default=None)  # type: Optional[Text]
    baseline_branch_name = attr.ib(default=None)  # type: Optional[Text]
    compare_with_parent_branch = attr.ib(default=None)  # type: Optional[bool]
    ignore_baseline = attr.ib(default=None)  # type: Optional[bool]
    save_failed_tests = attr.ib(default=None)  # type: Optional[bool]
    save_new_tests = attr.ib(default=None)  # type: Optional[bool]
    save_diffs = attr.ib(default=None)  # type: Optional[bool]
    dont_close_batches = attr.ib(default=None)  # type: Optional[bool]


@attr.s
class EyesCheckConfig(object):
    send_dom = attr.ib(default=None)  # type: Optional[bool]
    match_timeout = attr.ib(default=None)  # type: Optional[float]
    force_full_page_screenshot = attr.ib(default=None)  # type: Optional[bool]


@attr.s
class EyesClassicConfig(object):
    wait_before_screenshots = attr.ib(default=None)  # type: Optional[float]
    stitch_mode = attr.ib(default=None)  # type: Optional[StitchMode]
    hide_scrollbars = attr.ib(default=None)  # type: Optional[bool]
    hide_caret = attr.ib(default=None)  # type: Optional[bool]
    stitch_overlap = attr.ib(default=None)  # type: Optional[int]
    scroll_root_element = attr.ib(default=None)  # type: Optional[ElementReference]
    cut = attr.ib(default=None)  # type: Optional[Union[ImageCropRect,ImageCropRegion]]
    rotation = attr.ib(default=None)  # type: Optional[int]
    scale_ratio = attr.ib(default=None)  # type: Optional[float]


@attr.s
class EyesUFGConfig(object):
    concurrent_sessions = attr.ib(default=None)  # type: Optional[int]
    browsers_info = attr.ib(default=None)  # type: Optional[List[BrowserInfo]]
    visual_grid_options = attr.ib(default=None)  # type: Optional[Dict[Text, Any]]
    layout_breakpoints = attr.ib(default=None)  # type: Optional[Union[bool, List[int]]]
    disable_browser_fetching = attr.ib(default=None)  # type: Optional[bool]


@attr.s
class EyesConfig(
    EyesBaseConfig, EyesOpenConfig, EyesCheckConfig, EyesClassicConfig, EyesUFGConfig
):
    @classmethod
    def convert(cls, is_selenium, config):
        # type: (bool, Configuration) -> EyesConfig
        if config.cut_provider:
            cut = ImageCropRect(
                config.cut_provider.header,
                config.cut_provider.right,
                config.cut_provider.footer,
                config.cut_provider.left,
            )
        else:
            cut = None
        return cls(
            # EyesBaseConfig
            logs=None,  # TODO: verify
            debug_screenshots=DebugScreenshotHandler(
                config.save_debug_screenshots,
                config.debug_screenshots_path,
                config.debug_screenshots_prefix,
            ),
            agent_id=config.agent_id,
            api_key=config.api_key,
            server_url=config.server_url,
            proxy=Proxy.convert(config.proxy),
            is_disabled=config.is_disabled,
            connection_timeout=config._timeout,  # TODO: verify
            remove_session=None,  # TODO: verify
            remote_events=None,  # TODO: verify
            # EyesOpenConfig
            app_name=config.app_name,
            test_name=config.test_name,
            display_name=None,  # TODO: verify
            viewport_size=Size.convert_viewport(config.viewport_size),
            session_type=config.session_type,
            properties=CustomProperty.convert(config.properties),
            batch=Batch.convert(config.batch),
            default_match_settings=MatchSettings.convert(
                is_selenium, config.default_match_settings
            ),
            host_app=config.host_app,
            host_o_s=config.host_os,
            host_o_s_info=None,  # TODO: verify
            host_app_info=None,  # TODO: verify
            device_info=None,  # TODO: verify
            baseline_env_name=config.baseline_env_name,
            environment_name=config.environment_name,
            branch_name=config.branch_name,
            parent_branch_name=config.parent_branch_name,
            baseline_branch_name=config.baseline_branch_name,
            compare_with_parent_branch=None,  # TODO: verify
            ignore_baseline=None,  # TODO: verify
            save_failed_tests=config.save_failed_tests,
            save_new_tests=config.save_new_tests,
            save_diffs=config.save_diffs,
            dont_close_batches=True,  # TODO: verify
            # EyesCheckConfig
            send_dom=config.send_dom,
            match_timeout=config.match_timeout,
            force_full_page_screenshot=config.force_full_page_screenshot,
            # EyesClassicConfig
            wait_before_screenshots=config.wait_before_screenshots,
            stitch_mode=config.stitch_mode,
            hide_scrollbars=config.hide_scrollbars,
            hide_caret=config.hide_caret,
            stitch_overlap=config.stitch_overlap,
            scroll_root_element=None,  # TODO: verify
            cut=cut,
            rotation=config.rotation,
            scale_ratio=config.scale_ratio,
            # EyesUFGConfig
            concurrent_sessions=None,  # TODO: verify
            browsers_info=browsers_info_convert(config.browsers_info),
            visual_grid_options=record_convert(config.visual_grid_options),
            layout_breakpoints=config.layout_breakpoints,
            disable_browser_fetching=config.disable_browser_fetching,
        )


@attr.s
class ContextReference(object):
    frame = attr.ib()  # type: FrameReference
    scroll_root_element = attr.ib(default=None)  # type: Optional[ElementReference]

    @classmethod
    def convert(cls, is_selenium, frame_locators):
        # type: (bool, List[FrameLocator]) -> List[ContextReference]
        return [
            cls(
                frame=frame_reference_convert(
                    is_selenium,
                    frame_locator.frame_selector,
                    frame_locator.frame_element,
                    frame_locator.frame_index,
                    frame_locator.frame_name_or_id,
                ),
                scroll_root_element=element_reference_convert(
                    is_selenium,
                    frame_locator.scroll_root_selector,
                    frame_locator.scroll_root_element,
                ),
            )
            for frame_locator in frame_locators
        ]


@attr.s
class ScreenshotSettings(object):
    region = attr.ib(default=None)  # type: Optional[RegionReference]
    frames = attr.ib(default=None)  # type: Optional[List[ContextReference]]
    scroll_root_element = attr.ib(default=None)  # type: Optional[ElementReference]
    fully = attr.ib(default=None)  # type: Optional[bool]


@attr.s
class CheckSettingsHooks(object):
    before_capture_screenshot = attr.ib()  # type: Text


@attr.s
class CheckSettings(MatchSettings, ScreenshotSettings):
    name = attr.ib(default=None)  # type: Optional[Text]
    disable_browser_fetching = attr.ib(default=None)  # type: Optional[bool]
    layout_breakpoints = attr.ib(default=None)  # type: Optional[bool, List[int]]
    visual_grid_options = attr.ib(default=None)  # type: Optional[Dict[Text, Any]]
    hooks = attr.ib(default=None)  # type: Optional[CheckSettingsHooks]
    render_id = attr.ib(default=None)  # type: Optional[Text]
    variation_group_id = attr.ib(default=None)  # type: Optional[Text]
    timeout = attr.ib(default=None)  # type: Optional[int]

    @classmethod
    def convert(cls, is_selenium, values):
        # type: (bool, SeleniumCheckSettingsValues) -> CheckSettings
        if "beforeCaptureScreenshot" in values.script_hooks:
            hooks = CheckSettingsHooks(values.script_hooks["beforeCaptureScreenshot"])
        else:
            hooks = None
        return cls(
            # CheckSettings
            name=values.name,
            disable_browser_fetching=values.disable_browser_fetching,
            layout_breakpoints=values.layout_breakpoints,
            visual_grid_options=record_convert(values.visual_grid_options),
            hooks=hooks,
            render_id=None,  # TODO: verify
            variation_group_id=values.variation_group_id,
            timeout=values.timeout,
            # MatchSettings
            exact=None,  # TODO: verify
            match_level=values.match_level,
            send_dom=values.send_dom,
            use_dom=values.use_dom,
            enable_patterns=values.enable_patterns,
            ignore_caret=values.ignore_caret,
            ignore_displacements=values.ignore_displacements,
            accessibility_settings=None,  # TODO: verify
            ignore_regions=region_references_convert(
                is_selenium, values.ignore_regions
            ),
            layout_regions=region_references_convert(
                is_selenium, values.layout_regions
            ),
            strict_regions=region_references_convert(
                is_selenium, values.strict_regions
            ),
            content_regions=region_references_convert(
                is_selenium, values.content_regions
            ),
            floating_regions=floating_region_references_convert(
                is_selenium, values.floating_regions
            ),
            accessibility_regions=accessibility_region_references_convert(
                is_selenium, values.accessibility_regions
            ),
            # ScreenshotSettings
            region=target_reference_convert(is_selenium, values),
            frames=ContextReference.convert(is_selenium, values.frame_chain),
            scroll_root_element=element_reference_convert(
                is_selenium, values.scroll_root_selector, values.scroll_root_element
            ),
            fully=values.stitch_content,
        )


@attr.s
class LocateSettings(object):
    locator_names = attr.ib(factory=list)  # type: List[Text]
    first_only = attr.ib(default=None)  # type: Optional[bool]

    @classmethod
    def convert(cls, visual_locators_settings_values):
        # type: (VisualLocatorSettingsValues) -> LocateSettings
        return cls(
            visual_locators_settings_values.names,
            visual_locators_settings_values.first_only,
        )


@attr.s
class OCRSearchSettings(object):
    patterns = attr.ib(factory=list)  # type: List[Text]
    ignore_case = attr.ib(default=None)  # type: Optional[bool]
    first_only = attr.ib(default=None)  # type: Optional[bool]
    language = attr.ib(default=None)  # type: Optional[Text]

    @classmethod
    def convert(cls, search_settigns):
        # type: (TextRegionSettings) -> OCRSearchSettings
        return cls(
            search_settigns._patterns,
            search_settigns._ignore_case,
            search_settigns._first_only,
            search_settigns._language,
        )


@attr.s
class OCRExtractSettings(object):
    target = attr.ib()  # type: RegionReference
    hint = attr.ib(default=None)  # type: Optional[Text]
    min_match = attr.ib(default=None)  # type: Optional[float]
    language = attr.ib(default=None)  # type: Optional[Text]

    @classmethod
    def convert(cls, is_selenium, ocr_regions):
        # type: (bool, Tuple[OCRRegion]) -> List[OCRExtractSettings]
        return [
            cls(
                ocr_target_convert(is_selenium, region.target),
                region._hint,
                region._min_match,
                region._language,
            )
            for region in ocr_regions
        ]


@attr.s
class CloseBatchesSettings(object):
    batch_ids = attr.ib()  # type: List[Text]
    server_url = attr.ib()  # type: Optional[Text]
    api_key = attr.ib()  # type: Optional[Text]
    proxy = attr.ib()  # type: Optional[Proxy]

    @classmethod
    def convert(cls, enabled_batch_close):
        # type: (_EnabledBatchClose) -> CloseBatchesSettings
        return cls(
            list(enabled_batch_close._ids),
            enabled_batch_close.server_url,
            enabled_batch_close.api_key,
            Proxy.convert(enabled_batch_close.proxy),
        )


@attr.s
class DeleteTestSettings(object):
    test_id = attr.ib()  # type: Text
    batch_id = attr.ib()  # type: Text
    secret_token = attr.ib()  # type: Text
    server_url = attr.ib()  # type: Optional[Text]
    api_key = attr.ib()  # type: Optional[Text]
    proxy = attr.ib()  # type: Optional[Proxy]

    @classmethod
    def convert(cls, test_results):
        # type: (TestResults) -> DeleteTestSettings
        server_url, api_key, proxy = test_results._connection_config  # noqa
        return cls(
            test_results.id,
            test_results.batch_id,
            test_results.secret_token,
            server_url,
            api_key,
            Proxy.convert(proxy),
        )


@attr.s
class ServerInfo(object):
    logs_dir = attr.ib()  # type: Text


def marshal_webdriver_ref(driver):
    # type: (WebDriver) -> dict
    return {
        "sessionId": driver.session_id,
        "serverUrl": driver.command_executor._url,  # noqa
        "capabilities": driver.capabilities,
    }


def marshal_configuration(is_selenium, configuration):
    # type: (bool, Configuration) -> dict
    eyes_config = EyesConfig.convert(is_selenium, configuration)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(eyes_config))


def marshal_check_settings(is_selenium, check_settings):
    # type: (bool, SeleniumCheckSettings) -> dict
    check_settings = CheckSettings.convert(is_selenium, check_settings.values)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(check_settings))


def marshal_locate_settings(locate_settings):
    # type: (VisualLocatorSettings) -> dict
    locate_settings = LocateSettings.convert(locate_settings.values)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(locate_settings))


def marshal_ocr_search_settings(search_settings):
    # type: (TextRegionSettings) -> dict
    search_settings = OCRSearchSettings.convert(search_settings)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(search_settings))


def marshal_ocr_extract_settings(is_selenium, extract_settings):
    # type: (bool, Tuple[OCRRegion]) -> dict
    extract_settings = OCRExtractSettings.convert(is_selenium, extract_settings)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(extract_settings))


def marshal_viewport_size(viewport_size):
    # type: (ViewPort) -> dict
    size = Size.convert_viewport(viewport_size)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(size))


def marshal_enabled_batch_close(close_batches):
    # type: (_EnabledBatchClose) -> dict
    close_batches = CloseBatchesSettings.convert(close_batches)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(close_batches))


def marshal_delete_test_settings(test_results):
    # type: (TestResults) -> dict
    delete_settings = DeleteTestSettings.convert(test_results)
    return _keys_underscore_to_camel_remove_none(cattr.unstructure(delete_settings))


def demarshal_match_result(results_dict):
    # type: (dict) -> MatchResult
    return attr_from_json(dumps(results_dict), MatchResult)


def demarshal_locate_result(results):
    # type: (dict) -> Dict[Text, List[APIRegion]]
    return {
        locator_id: [APIRegion.from_(region) for region in regions] if regions else []
        for locator_id, regions in results.items()
    }


def demarshal_test_results(results_dict_list, config):
    # type: (List[dict], Optional[Configuration]) -> List[TestResults]
    results = [attr_from_json(dumps(r), TestResults) for r in results_dict_list]
    if config:
        for result in results:
            if result:  # in case of internal USDK failure, None result is observed
                result.set_connection_config(
                    config.server_url, config.api_key, config.proxy
                )
    return results


def demarshal_server_info(info_dict):
    # type: (dict) -> ServerInfo
    return ServerInfo(info_dict["logsDir"])


def _keys_underscore_to_camel_remove_none(obj):
    if isinstance(obj, dict):
        return {
            underscore_to_camelcase(k): _keys_underscore_to_camel_remove_none(v)
            for k, v in obj.items()
            if v is not None
        }
    elif isinstance(obj, list):
        return [_keys_underscore_to_camel_remove_none(i) for i in obj]
    else:
        return obj



from copy import deepcopy
from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union, overload

import attr
from six import raise_from

from applitools.common.config import Configuration as ConfigurationBase
from applitools.common.ultrafastgrid import (
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    IosDeviceInfo,
    IRenderBrowserInfo,
    RenderBrowserInfo,
    ScreenOrientation,
    VisualGridOption,
)
from applitools.common.utils import argument_guard
from applitools.common.utils.json_utils import JsonInclude
from applitools.common.validators import is_list_or_tuple

from .misc import BrowserType, StitchMode

if TYPE_CHECKING:
    from applitools.common.ultrafastgrid import DeviceName
    from applitools.core.cut import CutProvider

__all__ = ("Configuration",)

DEFAULT_WAIT_BEFORE_SCREENSHOTS_MS = 1000  # type: int  # ms


@attr.s
class Configuration(ConfigurationBase):
    force_full_page_screenshot = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: bool
    wait_before_screenshots = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        default=DEFAULT_WAIT_BEFORE_SCREENSHOTS_MS,
    )  # type: int  # ms
    stitch_mode = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=StitchMode.Scroll
    )  # type: StitchMode
    hide_scrollbars = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: bool
    hide_caret = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=False
    )  # type: bool
    # Indicates that a mobile simulator is being used
    is_simulator = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[bool]

    # Rendering Configuration
    _browsers_info = attr.ib(
        metadata={JsonInclude.NON_NONE: "browsersInfo"}, init=False, factory=list
    )  # type: List[IRenderBrowserInfo]
    visual_grid_options = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Tuple[VisualGridOption]]
    disable_browser_fetching = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=True
    )  # type: bool
    enable_cross_origin_rendering = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=True
    )  # type: bool
    dont_use_cookies = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=False
    )  # type: bool
    layout_breakpoints = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Union[bool, List[int]]]
    scale_ratio = attr.ib(default=None)  # type: Optional[float]
    cut_provider = attr.ib(default=None)  # type: Optional[CutProvider]
    rotation = attr.ib(default=None)  # type: Optional[int]

    def set_force_full_page_screenshot(self, force_full_page_screenshot):
        # type: (bool) -> Configuration
        self.force_full_page_screenshot = force_full_page_screenshot
        return self

    def set_wait_before_screenshots(self, wait_before_screenshots):
        # type: (int) -> Configuration
        self.wait_before_screenshots = wait_before_screenshots
        return self

    def set_stitch_mode(self, stitch_mode):
        # type: (StitchMode) -> Configuration
        self.stitch_mode = stitch_mode
        return self

    def set_hide_scrollbars(self, hide_scrollbars):
        # type: (bool) -> Configuration
        self.hide_scrollbars = hide_scrollbars
        return self

    def set_hide_caret(self, hide_caret):
        # type: (bool) -> Configuration
        self.hide_caret = hide_caret
        return self

    def set_visual_grid_options(self, *options):
        # type: (*VisualGridOption) -> Configuration
        if options == (None,):
            self.visual_grid_options = None
        else:
            argument_guard.are_(options, VisualGridOption)
            self.visual_grid_options = options
        return self

    def set_disable_browser_fetching(self, disable_browser_fetching):
        # type: (bool) -> Configuration
        self.disable_browser_fetching = disable_browser_fetching
        return self

    def set_enable_cross_origin_rendering(self, enable_cross_origin_rendering):
        # type: (bool) -> Configuration
        self.enable_cross_origin_rendering = enable_cross_origin_rendering
        return self

    def set_dont_use_cookies(self, dont_use_cookies):
        # type: (bool) -> Configuration
        self.dont_use_cookies = dont_use_cookies
        return self

    @overload
    def set_layout_breakpoints(self, enabled):
        # type: (bool) -> Configuration
        pass

    @overload
    def set_layout_breakpoints(self, *breakpoints):
        # type: (*int) -> Configuration
        pass

    def set_layout_breakpoints(self, enabled_or_first, *rest):
        if isinstance(enabled_or_first, bool):
            self.layout_breakpoints = enabled_or_first
        elif isinstance(enabled_or_first, int):
            self.layout_breakpoints = [enabled_or_first] + list(rest)
        else:
            raise TypeError(
                "{} is not an instance of bool or int".format(enabled_or_first)
            )
        return self

    @overload  # noqa
    def add_browser(self, desktop_browser_info):
        # type: (DesktopBrowserInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, ios_device_info):
        # type: (IosDeviceInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, chrome_emulation_info):
        # type: (ChromeEmulationInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, width, height, browser_type):
        # type: (int, int, BrowserType) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, width, height, browser_type, baseline_env_name):
        # type: (int, int, BrowserType, Text) -> Configuration
        pass

    def add_browser(self, *args):  # noqa
        if isinstance(args[0], IRenderBrowserInfo):
            self._browsers_info.append(args[0])
        elif (
            isinstance(args[0], int)
            and isinstance(args[1], int)
            and isinstance(args[2], BrowserType)
        ):
            if len(args) == 4:
                baseline_env_name = args[3]
            else:
                baseline_env_name = self.baseline_env_name
            self._browsers_info.append(
                DesktopBrowserInfo(args[0], args[1], args[2], baseline_env_name)
            )
        else:
            raise TypeError(
                "Unsupported parameter: \n\ttype: {} \n\tvalue: {}".format(
                    type(args), args
                )
            )
        return self

    @overload
    def add_browsers(self, renders_info):
        # type:(List[Union[DesktopBrowserInfo,IosDeviceInfo,ChromeEmulationInfo]])->Configuration  # noqa
        pass

    @overload
    def add_browsers(self, renders_info):
        # type:(*Union[DesktopBrowserInfo,IosDeviceInfo,ChromeEmulationInfo])->Configuration
        pass

    def add_browsers(self, *renders_info):
        if len(renders_info) == 1 and is_list_or_tuple(renders_info[0]):
            renders_info = renders_info[0]

        for render_info in renders_info:
            try:
                self.add_browser(render_info)
            except TypeError as e:
                raise_from(TypeError("Wrong argument in .add_browsers()"), e)
        return self

    def add_device_emulation(self, device_name, orientation=ScreenOrientation.PORTRAIT):
        # type: (DeviceName, ScreenOrientation) -> Configuration
        argument_guard.not_none(device_name)
        self.add_browser(ChromeEmulationInfo(device_name, orientation))
        return self

    @property
    def browsers_info(self):
        # type: () -> List[RenderBrowserInfo]
        if self._browsers_info:
            return self._browsers_info
        if self.viewport_size:
            browser_info = DesktopBrowserInfo(
                width=self.viewport_size["width"],
                height=self.viewport_size["height"],
                browser_type=BrowserType.CHROME,
                baseline_env_name=self.baseline_env_name,
            )
            return [browser_info]
        return []

    def clone(self):
        # type: () -> Configuration
        # TODO: Remove this huck when get rid of Python2
        conf = super(Configuration, self).clone()
        conf._browsers_info = deepcopy(self._browsers_info)
        conf.visual_grid_options = deepcopy(self.visual_grid_options)
        return conf



from __future__ import absolute_import

from typing import TYPE_CHECKING, overload

from .selenium_check_settings import SeleniumCheckSettings

if TYPE_CHECKING:
    from applitools.common import Region
    from applitools.common.utils.custom_types import (
        AnyWebElement,
        BySelector,
        CssSelector,
        FrameIndex,
        FrameNameOrId,
    )

    from .target_path import Locator

__all__ = ("Target",)


class Target(object):
    """
    Target for an eyes.check_window/region.
    """

    @staticmethod
    def window():
        # type: () -> SeleniumCheckSettings
        return SeleniumCheckSettings()

    @staticmethod  # noqa
    @overload
    def region(region):
        # type: (Region) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(css_selector):
        # type: (CssSelector) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(by_selector):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(target_path):
        # type: (Locator) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    def region(region):
        return SeleniumCheckSettings().region(region)

    @staticmethod  # noqa
    @overload
    def frame(frame_name_or_id):
        # type: (FrameNameOrId) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def frame(frame_element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def frame(frame_index):
        # type: (FrameIndex) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def frame(frame_by_selector):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def frame(target_path):
        # type: (Locator) -> SeleniumCheckSettings
        pass

    @staticmethod  # noqa
    def frame(frame):
        return SeleniumCheckSettings().frame(frame)



from applitools.common import (
    DeviceName,
    FileLogger,
    MatchLevel,
    RectangleSize,
    Region,
    StdoutLogger,
    TestResultContainer,
    TestResults,
    TestResultsSummary,
    logger,
)
from applitools.common.accessibility import (  # noqa
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
    AccessibilitySettings,
)
from applitools.common.config import BatchInfo  # noqa
from applitools.common.geometry import AccessibilityRegion
from applitools.common.selenium import BrowserType, Configuration, StitchMode  # noqa
from applitools.common.server import FailureReports  # noqa
from applitools.common.ultrafastgrid import (  # noqa
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    IosDeviceInfo,
    IosDeviceName,
    IosVersion,
    ScreenOrientation,
)
from applitools.core.batch_close import BatchClose  # noqa
from applitools.core.cut import (  # noqa
    FixedCutProvider,
    NullCutProvider,
    UnscaledFixedCutProvider,
)
from applitools.core.extract_text import OCRRegion, TextRegionSettings
from applitools.core.fluent.region import AccessibilityRegionByRectangle  # noqa

from .eyes import Eyes
from .fluent.target import Target  # noqa
from .fluent.target_path import TargetPath
from .runner import ClassicRunner, RunnerOptions, VisualGridRunner

__all__ = (
    # noqa
    "BatchInfo",
    "Region",
    "MatchLevel",
    "logger",
    "StdoutLogger",
    "FileLogger",
    "Eyes",
    "Target",
    "TargetPath",
    "FailureReports",
    "StitchMode",
    "VisualGridRunner",
    "RunnerOptions",
    "BrowserType",
    "DeviceName",
    "Configuration",
    "ScreenOrientation",
    "FixedCutProvider",
    "NullCutProvider",
    "UnscaledFixedCutProvider",
    "ClassicRunner",
    "RectangleSize",
    "TestResults",
    "TestResultContainer",
    "TestResultsSummary",
    "BatchClose",
    "AccessibilityRegionType",
    "AccessibilityLevel",
    "AccessibilitySettings",
    "AccessibilityGuidelinesVersion",
    "AccessibilityRegionByRectangle",
    "AccessibilityRegion",
    "IosDeviceName",
    "IosDeviceInfo",
    "IosVersion",
    "ChromeEmulationInfo",
    "DesktopBrowserInfo",
    "OCRRegion",
    "TextRegionSettings",
)



from typing import List, Optional, Text, Union

import attr

from applitools.common import ProxySettings
from applitools.common.config import DEFAULT_SERVER_URL
from applitools.common.utils import argument_guard
from applitools.common.utils.general_utils import get_env_with_prefix


@attr.s
class _EnabledBatchClose(object):
    _ids = attr.ib()  # type: List[Text]
    server_url = attr.ib()  # type: Text
    api_key = attr.ib()  # type: Text
    proxy = attr.ib(default=None)  # type: Optional[ProxySettings]

    def set_url(self, url):
        # type: (Text) -> _EnabledBatchClose
        self.server_url = url
        return self

    def set_api_key(self, api_key):
        # type: (Text) -> _EnabledBatchClose
        self.api_key = api_key
        return self

    def set_proxy(self, proxy):
        # type: (ProxySettings) -> _EnabledBatchClose
        argument_guard.is_a(proxy, ProxySettings)
        self.proxy = proxy
        return self

    def close(self):
        from applitools.selenium.__version__ import __version__
        from applitools.selenium.command_executor import CommandExecutor
        from applitools.selenium.eyes import EyesRunner
        from applitools.selenium.universal_sdk_types import marshal_enabled_batch_close

        cmd = CommandExecutor.get_instance(EyesRunner.BASE_AGENT_ID, __version__)
        marshaled = marshal_enabled_batch_close(self)
        cmd.core_close_batches(marshaled)


@attr.s
class BatchClose(object):
    api_key = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_API_KEY", None)
    )  # type: Optional[Text]
    server_url = attr.ib(default=DEFAULT_SERVER_URL)  # type: Text
    proxy = attr.ib(default=None)  # type: Optional[ProxySettings]

    def set_url(self, url):
        # type: (Text) -> BatchClose
        self.server_url = url
        return self

    def set_api_key(self, api_key):
        # type: (Text) -> BatchClose
        self.api_key = api_key
        return self

    def set_proxy(self, proxy):
        # type: (ProxySettings) -> BatchClose
        argument_guard.is_a(proxy, ProxySettings)
        self.proxy = proxy
        return self

    def set_batch_ids(self, *ids):
        # type: (Union[Text, List[Text]]) -> _EnabledBatchClose
        if isinstance(ids[0], list):
            ids = ids[0]
        return _EnabledBatchClose(ids, self.server_url, self.api_key, self.proxy)



from six import iteritems

from applitools.common.validators import is_list_or_tuple


def get_name_from_param(param):
    for var_name, var in iteritems(locals()):
        if var is param:
            return var_name


def not_none(param, exc_info=None):
    """
    Fails if the input parameter is None
    """
    if param is None:
        if exc_info:
            raise exc_info
        raise ValueError("{} is None".format(get_name_from_param(param)))


def not_list_or_tuple(param, exc_info=None):
    """
    Fails if param is not a list or tuple
    """
    if not is_list_or_tuple(param):
        if exc_info:
            raise exc_info
        raise ValueError("{} is not list or tuple".format(get_name_from_param(param)))


def is_valid_type(is_valid, message):
    """
    Fails if is_valid is false.
    """
    if not is_valid:
        raise TypeError(message)


def is_a(param, klass, exc_info=None):
    if not isinstance(param, klass):
        if exc_info:
            raise exc_info
        raise ValueError("`{}` is not instance of `{}`".format(param, klass))


def are_(params, klass, exc_info=None):
    for param in params:
        is_a(param, klass, exc_info)



import atexit
import logging
import weakref
from concurrent.futures import Future
from json import dumps, loads
from threading import Thread
from typing import Optional, Text
from uuid import uuid1

from websocket import WebSocket

from applitools.eyes_universal import get_instance, server

_all_sockets = []
_logger = logging.getLogger(__name__)


class USDKConnection(object):
    def __init__(self, websocket, server_log_file=None):
        # type: (WebSocket, Optional[Text]) -> None
        self.server_log_file = server_log_file
        self._websocket = websocket
        self._response_futures = {}
        weak_socket = weakref.ref(self._websocket)
        self._receiver_thread = Thread(
            target=self._receiver_loop,
            name="USDK Receiver",
            args=(weak_socket, self._response_futures),
        )
        # Receiver threads are designed to exit even if they serve leaked unclosed
        # connections. But non-daemon threads are joined on shutdown deadlocking with
        # garbage collector and atexit manager that should close the socket
        self._receiver_thread.daemon = True
        _all_sockets.append(weak_socket)
        self._receiver_thread.start()

    @classmethod
    def create(cls, server=None):
        # type: (Optional[server.SDKServer]) -> USDKConnection
        server = server or get_instance()
        websocket = WebSocket()
        websocket.connect("ws://localhost:{}/eyes".format(server.port))
        websocket.settimeout(20 * 60)
        return cls(websocket, server.log_file_name)

    def notification(self, name, payload):
        # type: (Text, dict) -> None
        self._websocket.send(dumps({"name": name, "payload": payload}))

    def command(self, name, payload, wait_result, wait_timeout):
        # type: (Text, dict, bool, float) -> Optional[dict]
        key = str(uuid1())
        future = Future()
        self._response_futures[key] = future
        self._websocket.send(dumps({"name": name, "key": key, "payload": payload}))
        if wait_result:
            return future.result(wait_timeout)
        else:
            return None

    def close(self):
        if self._websocket:
            self._websocket.close()
            self._websocket = None
            self._receiver_thread = None

    def __del__(self):
        self.close()

    @staticmethod
    def _receiver_loop(weak_socket, response_futures):
        while True:
            try:
                socket = weak_socket()
                if not socket:
                    raise EOFError
                response = socket.recv()
                del socket
                if not response:
                    raise EOFError
                response = loads(response)
                if "key" in response:
                    future = response_futures.pop(response["key"])
                    future.set_result(response)
                else:
                    _logger.warning("Unexpected server-initiated event %s", response)
            except Exception as exc:
                for future in response_futures.values():
                    future.set_exception(exc)
                break


@atexit.register
def ensure_all_closed():
    for weak_ref in _all_sockets:
        socket = weak_ref()
        if socket:
            socket.close()



from __future__ import absolute_import, print_function, unicode_literals

import typing
from concurrent.futures import TimeoutError

from applitools.common import (
    DiffsFoundError,
    EyesError,
    NewTestError,
    TestFailedError,
    TestResultContainer,
    TestResultsSummary,
)
from applitools.common.config import DEFAULT_ALL_TEST_RESULTS_TIMEOUT

from .__version__ import __version__
from .command_executor import CommandExecutor, ManagerType
from .universal_sdk_types import demarshal_server_info, demarshal_test_results

if typing.TYPE_CHECKING:
    from typing import Optional, Union


class EyesRunner(object):
    AUTO_CLOSE_MODE_SYNC = True
    BASE_AGENT_ID = "eyes.sdk.python"

    def __init__(self, manager_type, concurrency=None, is_legacy=None):
        # type: (ManagerType, Optional[int], Optional[bool]) -> None
        self._commands = CommandExecutor.get_instance(self.BASE_AGENT_ID, __version__)
        self._ref = self._commands.core_make_manager(
            manager_type, concurrency, is_legacy
        )

    @classmethod
    def get_server_info(cls):
        cmd = CommandExecutor.get_instance(cls.BASE_AGENT_ID, __version__)
        result = cmd.server_get_info()
        return demarshal_server_info(result)

    def get_all_test_results(
        self, should_raise_exception=True, timeout=DEFAULT_ALL_TEST_RESULTS_TIMEOUT
    ):
        # type: (bool, Optional[int]) -> TestResultsSummary
        try:
            results = self._commands.manager_close_all_eyes(self._ref, timeout)
        except TimeoutError:
            raise EyesError("Tests didn't finish in {} seconds".format(timeout))
        # We don't have server_url, api_key and proxy settings in runner
        # USDK should return them back as a part of TestResults
        structured_results = demarshal_test_results(results, None)
        for r in structured_results:
            log_session_results_and_raise_exception(should_raise_exception, r)
        return TestResultsSummary(
            [TestResultContainer(result, None, None) for result in structured_results]
        )


class RunnerOptions(object):
    concurrency = 5

    def test_concurrency(self, value):
        # type: (int) -> RunnerOptions
        self.concurrency = value
        return self


class VisualGridRunner(EyesRunner):
    AUTO_CLOSE_MODE_SYNC = False

    def __init__(self, options_or_concurrency=RunnerOptions()):
        # type: (Union[RunnerOptions, int]) -> None
        if isinstance(options_or_concurrency, int):
            concurrency = options_or_concurrency * 5  # legacy factor
            is_legacy = True
        else:
            concurrency = options_or_concurrency.concurrency
            is_legacy = False
        super(VisualGridRunner, self).__init__(ManagerType.VG, concurrency, is_legacy)


class ClassicRunner(EyesRunner):
    def __init__(self):
        super(ClassicRunner, self).__init__(ManagerType.CLASSIC)


def log_session_results_and_raise_exception(raise_ex, results):
    results_url = results.url
    scenario_id_or_name = results.name
    app_id_or_name = results.app_name
    if results.steps == 0:
        print("--- Test has no checks. \n\tSee details at ", results_url)
        if raise_ex:
            raise TestFailedError(results, scenario_id_or_name, app_id_or_name)
    elif results.is_unresolved:
        if results.is_new:
            print(
                "--- New test ended. \n\tPlease approve the new baseline at",
                results_url,
            )
            if raise_ex:
                raise NewTestError(results, scenario_id_or_name, app_id_or_name)
        else:
            print("--- Differences are found. \n\tSee details at", results_url)
            if raise_ex:
                raise DiffsFoundError(results, scenario_id_or_name, app_id_or_name)
    elif results.is_failed:
        print("--- Failed test ended. \n\tSee details at", results_url)
        if raise_ex:
            raise TestFailedError(results, scenario_id_or_name, app_id_or_name)
    else:
        print("--- Test passed. \n\tSee details at", results_url)



from enum import Enum
from typing import Any, Text


class ScreenOrientation(Enum):
    PORTRAIT = "portrait"
    LANDSCAPE = "landscape"


class DeviceName(Enum):
    Blackberry_PlayBook = "Blackberry PlayBook"
    BlackBerry_Z30 = "BlackBerry Z30"
    Galaxy_A5 = "Galaxy A5"
    Galaxy_Note_10 = "Galaxy Note 10"
    Galaxy_Note_10_Plus = "Galaxy Note 10 Plus"
    Galaxy_Note_2 = "Galaxy Note 2"
    Galaxy_Note_3 = "Galaxy Note 3"
    Galaxy_Note_4 = "Galaxy Note 4"
    Galaxy_Note_8 = "Galaxy Note 8"
    Galaxy_Note_9 = "Galaxy Note 9"
    Galaxy_S3 = "Galaxy S3"
    Galaxy_S5 = "Galaxy S5"
    Galaxy_S8 = "Galaxy S8"
    Galaxy_S8_Plus = "Galaxy S8 Plus"
    Galaxy_S9 = "Galaxy S9"
    Galaxy_S9_Plus = "Galaxy S9 Plus"
    Galaxy_S10 = "Galaxy S10"
    Galaxy_S10_Plus = "Galaxy S10 Plus"
    Galaxy_S20 = "Galaxy S20"
    iPad = "iPad"
    iPad_6th_Gen = "iPad 6th Gen"
    iPad_7th_Gen = "iPad 7th Gen"
    iPad_Air_2 = "iPad Air 2"
    iPad_Mini = "iPad Mini"
    iPad_Pro = "iPad Pro"
    iPhone_11 = "iPhone 11"
    iPhone_11_Pro = "iPhone 11 Pro"
    iPhone_11_Pro_Max = "iPhone 11 Pro Max"
    iPhone_4 = "iPhone 4"
    iPhone_5SE = "iPhone 5/SE"
    iPhone_6_7_8 = "iPhone 6/7/8"
    iPhone_6_7_8_Plus = "iPhone 6/7/8 Plus"
    iPhone_X = "iPhone X"
    iPhone_XR = "iPhone XR"
    iPhone_XS = "iPhone XS"
    iPhone_XS_Max = "iPhone XS Max"
    Kindle_Fire_HDX = "Kindle Fire HDX"
    Laptop_with_HiDPI_screen = "Laptop with HiDPI screen"
    Laptop_with_MDPI_screen = "Laptop with MDPI screen"
    Laptop_with_touch = "Laptop with touch"
    LG_G6 = "LG G6"
    LG_Optimus_L70 = "LG Optimus L70"
    Microsoft_Lumia_550 = "Microsoft Lumia 550"
    Microsoft_Lumia_950 = "Microsoft Lumia 950"
    Nexus_10 = "Nexus 10"
    Nexus_4 = "Nexus 4"
    Nexus_5 = "Nexus 5"
    Nexus_5X = "Nexus 5X"
    Nexus_6 = "Nexus 6"
    Nexus_6P = "Nexus 6P"
    Nexus_7 = "Nexus 7"
    Nokia_Lumia_520 = "Nokia Lumia 520"
    Nokia_N9 = "Nokia N9"
    OnePlus_7T = "OnePlus 7T"
    OnePlus_7T_Pro = "OnePlus 7T Pro"
    # OnePlus_8 = 'OnePlus 8'
    # OnePlus_8_Pro = 'OnePlus 8 Pro'
    Pixel_2 = "Pixel 2"
    Pixel_2_XL = "Pixel 2 XL"
    Pixel_3 = "Pixel 3"
    Pixel_3_XL = "Pixel 3 XL"
    Pixel_4 = "Pixel 4"
    Pixel_4_XL = "Pixel 4 XL"
    Pixel_5 = "Pixel 5"


class IosVersion(Enum):
    LATEST = "latest"
    ONE_VERSION_BACK = "latest-1"


class IosDeviceName(Enum):
    iPhone_13_Pro_Max = "iPhone 13 Pro Max"
    iPhone_13_Pro = "iPhone 13 Pro"
    iPhone_13 = "iPhone 13"
    iPhone_12_Pro_Max = "iPhone 12 Pro Max"
    iPhone_12_Pro = "iPhone 12 Pro"
    iPhone_12 = "iPhone 12"
    iPhone_12_mini = "iPhone 12 mini"
    iPhone_11_Pro = "iPhone 11 Pro"
    iPhone_11_Pro_Max = "iPhone 11 Pro Max"
    iPhone_11 = "iPhone 11"
    iPhone_XR = "iPhone XR"
    iPhone_XS = "iPhone Xs"
    iPhone_X = "iPhone X"
    iPhone_8 = "iPhone 8"
    iPhone_7 = "iPhone 7"
    iPad_Pro_3 = "iPad Pro (12.9-inch) (3rd generation)"
    iPad_7 = "iPad (7th generation)"
    iPad_9 = "iPad (9th generation)"
    iPad_Air_2 = "iPad Air (2nd generation)"


class VisualGridOption(object):
    def __init__(self, key, value):
        # type: (Text, Any) -> None
        self.key = key
        self.value = value

    def __eq__(self, other):
        # type: (VisualGridOption) -> bool
        return other and self.key == other.key and self.value == other.value



from __future__ import absolute_import

from appium.webdriver import WebElement as AppiumWebElement
from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement
from selenium.webdriver.support.event_firing_webdriver import EventFiringWebElement


def is_list_or_tuple(elm):
    return isinstance(elm, (list, tuple))


def is_webelement(elm):

    return isinstance(
        elm,
        (SeleniumWebElement, AppiumWebElement, EventFiringWebElement),
    )



import codecs
import re
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))


def read(filename):
    return codecs.open(path.join(here, filename), "r", "utf-8").read()


def get_version():
    version_file = read("applitools/selenium/__version__.py")
    try:
        version = re.findall(r"^__version__ = \"([^']+)\"\r?$", version_file, re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

    return version


setup(version=get_version())



from __future__ import absolute_import

from typing import Text


def round_converter(x):
    # type: (float) -> int
    return int(round(x))


def str2bool(v):
    # type: (Text) -> bool
    if v is None:
        return False
    return v.lower() in ("yes", "true", "t", "1")



from .batch_close import BatchClose
from .cut import FixedCutProvider, NullCutProvider, UnscaledFixedCutProvider
from .extract_text import TextRegionSettings
from .feature import Feature
from .fluent import (
    CheckSettings,
    CheckSettingsValues,
    FloatingRegionByRectangle,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)
from .locators import VisualLocator, VisualLocatorSettings
from .triggers import MouseTrigger, TextTrigger

__all__ = (
    "TextTrigger",
    "MouseTrigger",
    "Feature",
    "FixedCutProvider",
    "NullCutProvider",
    "UnscaledFixedCutProvider",
    "CheckSettings",
    "CheckSettingsValues",
    "GetRegion",
    "GetFloatingRegion",
    "FloatingRegionByRectangle",
    "RegionByRectangle",
    "BatchClose",
    "VisualLocator",
    "VisualLocatorSettings",
    "TextRegionSettings",
)



from typing import TYPE_CHECKING

from selenium.webdriver.common.by import By

if TYPE_CHECKING:
    pass


class Locator(object):
    def __init__(self, parent, selector_or_by, selector=None):
        self.parent = parent
        if selector is None:
            self.by = By.CSS_SELECTOR if selector_or_by else None
            self.selector = selector_or_by
        else:
            self.by = selector_or_by
            self.selector = selector

    def __eq__(self, other):
        return type(self) is type(other) and vars(self) == vars(other)

    def __repr__(self):
        parent = repr(self.parent) if self.parent else "TargetPath"
        if self.by == By.CSS_SELECTOR:
            return parent + ".{}({!r})".format(
                self.FACTORY_METHOD, self.selector  # noqa
            )
        else:
            return parent + ".{}({}, {!r})".format(
                self.FACTORY_METHOD, _by_repr(self.by), self.selector  # noqa
            )


class TargetPath(object):
    @staticmethod
    def region(selector_or_by, selector=None):
        return RegionLocator(None, selector_or_by, selector)

    @staticmethod
    def frame(number_or_id_or_name_or_by, selector=None):
        return FrameLocator(None, number_or_id_or_name_or_by, selector)

    @staticmethod
    def shadow(selector_or_by, selector=None):
        return ShadowDomLocator(None, selector_or_by, selector)


class ShadowDomLocator(Locator):
    FACTORY_METHOD = "shadow"

    def region(self, selector_or_by, selector=None):
        return RegionLocator(self, selector_or_by, selector)

    def frame(self, number_or_id_or_name_or_by, selector=None):
        return FrameLocator(self, number_or_id_or_name_or_by, selector)

    def shadow(self, selector_or_by, selector=None):
        return ShadowDomLocator(self, selector_or_by, selector)


class RegionLocator(Locator):
    FACTORY_METHOD = "region"


class FrameLocator(Locator):
    FACTORY_METHOD = "frame"

    def __init__(self, parent, number_or_id_or_name_or_by, selector=None):
        if selector is None:
            super(FrameLocator, self).__init__(parent, None, None)
            self.number_or_id_or_name = number_or_id_or_name_or_by
        else:
            super(FrameLocator, self).__init__(
                parent, number_or_id_or_name_or_by, selector
            )
            self.number_or_id_or_name = None

    def __repr__(self):
        parent = repr(self.parent) if self.parent else "TargetPath"
        if self.number_or_id_or_name is not None:
            return parent + ".{}({!r})".format(
                self.FACTORY_METHOD, self.number_or_id_or_name
            )
        else:
            return parent + ".{}({}, {!r})".format(
                self.FACTORY_METHOD, _by_repr(self.by), self.selector
            )

    def region(self, selector_or_by, selector=None):
        return RegionLocator(self, selector_or_by, selector)

    def frame(self, number_or_id_or_name_or_by, selector=None):
        return FrameLocator(self, number_or_id_or_name_or_by, selector)

    def shadow(self, selector_or_by, selector=None):
        return ShadowDomLocator(self, selector_or_by, selector)


def _by_repr(by):
    return "By." + by.upper().replace(" ", "_")
